@@ -1,6 +1,7 @@
-#!/usr/bin/perl -w
+package XML::XPathScript;
 
 use strict;
+use warnings;
 
 =pod "
 
@@ -79,9 +80,8 @@ is to be invoked like this:
 
    xpathscript mydocument.xml mystylesheet.xps
 
-It will produce the resulting document on standard output. More
-options will be added later (select output file, handle multiple
-output files, pass parameters to the stylesheet etc.).
+It will produce the resulting document on standard output. For more
+options, refer to xpathscript's man page.
 
 =head2 Functions and global variables available in the stylesheet
 
@@ -96,413 +96,6 @@ all code written in the embedded stylesheet dialect.
 
 =cut "
 
-package XML::XPathScript::Toys;
-
-use XML::XPath::Node;
-
-use vars '@ISA', '@EXPORT';
-use Exporter;
-@ISA = ('Exporter');
-@EXPORT = qw(
-        findnodes
-        findvalue
-        findvalues
-        findnodes_as_string
-        apply_templates
-	call_template
-        matches
-        set_namespace
-		DO_SELF_AND_KIDS
-		DO_SELF_ONLY
-		DO_NOT_PROCESS
-        );
-
-=pod "
-
-=item I<DO_SELF_AND_KIDS>
-
-=item I<DO_SELF_ONLY>
-
-=item I<DO_NOT_PROCESS>
-
-Symbolic constants evaluating respectively to 1, -1 and 0, to be used
-as mnemotechnic return values in C<testcode> routines instead of the
-numeric values which are harder to remember.
-
-=cut "
-
-    sub DO_SELF_AND_KIDS () { return  1 }
-    sub DO_SELF_ONLY     () { return -1 }
-    sub DO_NOT_PROCESS   () { return  0 }
-
-# quieten warnings when compiling this module
-sub apply_templates (;$@);
-
-=pod "
-
-=item I<findnodes($path)>
-
-=item I<findnodes($path, $context)>
-
-Returns a list of nodes found by XPath expression $path, optionally
-using $context as the context node (default is the root node of the
-current document).  In scalar context returns a NodeSet object.
-
-=cut "
-
-sub findnodes {
-    $XML::XPathScript::current->{xp}->findnodes(@_);
-}
-
-=pod "
-
-=item  I<findvalue($path)>
-
-=item  I<findvalue($path, $context)>
-
-Evaluates XPath expression $path and returns the result, as either a
-"XML::XPath::Literal", a "XML::XPath::Boolean" or a
-"XML::XPath::Number" object. If the path returns a NodeSet,
-$nodeset->to_literal is called automatically for you (and thus a
-"XML::XPath::Literal" is returned). Note that for each of the objects
-stringification is overloaded, so you can just print the value found,
-or manipulate it in the ways you would a normal perl value (e.g. using
-regular expressions) - just beware that the result of such
-stringification will be UTF8-encoded (see L<perlunicode>), as just
-about everything under the XML sun is.
-
-=cut "
-
-sub findvalue {
-    $XML::XPathScript::current->{xp}->findvalue(@_);
-}
-
-=pod "
-
-
-=item  I<findvalues($path)>
-
-=item  I<findvalues($path, $context)>
-
-Evaluates XPath expression $path as a nodeset expression, just like
-L</findnodes> would, but returns a list of UTF8-encoded XML strings
-instead of node objects.
-
-=cut "
-
-sub findvalues {
-    my @nodes = findnodes(@_);
-    map { findvalue('.', $_) } @nodes;
-}
-
-=pod "
-
-=item I<findnodes_as_string($path)>
-
-=item I<findnodes_as_string($path, $context)>
-
-Similar to L</findvalues> but concatenates the XML snippets.  The
-result is not guaranteed to be valid XML though.
-
-=cut "
-
-sub findnodes_as_string {
-    $XML::XPathScript::current->{xp}->findnodes_as_string(@_);
-}
-
-=pod "
-
-=item I<matches($node, $path)>
-
-=item I<matches($node, $path, $context)>
-
-Returns true if the node matches the path (optionally in context $context)
-
-=cut "
-
-sub matches {
-    $XML::XPathScript::current->{xp}->matches(@_);
-}
-
-sub set_namespace {
-    eval {
-        $XML::XPathScript::current->{xp}->set_namespace(@_);
-    };
-    if ($@) {
-        warn "set_namespace failed: $@";
-    }
-}
-
-=pod "
-
-=item I<apply_templates()>
-
-=item I<apply_templates($xpath)>
-
-=item I<apply_templates($xpath, $context)>
-
-=item I<apply_templates(@nodes)>
-
-This is where the whole magic in XPathScript resides: recursively
-applies the stylesheet templates to the nodes provided either
-literally (last invocation form) or through an XPath expression
-(second and third invocation forms), and returns a string
-concatenation of all results. If called without arguments at all,
-renders the whole document.
-
-Calls to I<apply_templates()> may occur both implicitly (at the top of
-the document, and for rendering subnodes when the templates choose not
-to handle that by themselves), and explicitly (from C<testcode>
-routines).
-
-If appropriate care is taken in all templates (especially the
-C<testcode> routines and the I<text()> template), the string result of
-I<apply_templates> need not be UTF-8 (see L<perlunicode>): it is thus
-possible using XPathScript to produce output in any character set
-without an extra translation pass.
-
-=cut "
-
-sub apply_templates (;$@) {
-    unless (@_) {
-        return apply_templates(findnodes('/'));
-    }
-
-    my ($arg1, @args) = @_;
-
-    if (!ref($arg1)) {
-        # called with a path to find
-#         warn "apply_templates with path '$arg1'\n";
-		if( my $nodes = findnodes($arg1, @args) )
-		{
-        	return apply_templates($nodes);
-		}
-		else{ return }
-    }
-
-    my $retval = '';
-    if ($arg1->isa('XML::XPath::NodeSet')) {
-        foreach my $node ($arg1->get_nodelist) {
-            $retval .= translate_node($node);
-        }
-    }
-    else {
-        $retval .= translate_node($arg1);
-        foreach my $node (@args) {
-            $retval .= translate_node($node);
-        }
-    }
-
-    return $retval;
-}
-
-=pod "
-
-=item I<call_template($node, $t, $templatename)>
-
-B<EXPERIMENTAL> - allows C<testcode> routines to invoke a template by
-name, even if the selectors do not fit (e.g. one can apply template B
-to an element node of type A). Returns the stylesheeted string
-computed out of $node just like L</apply_templates> would.
-
-=cut "
-
-sub call_template {
-    my ($self,$t,$template)=@_;
-
-    if (defined(my $sub=$template->{testcode})) {
-	return &$sub($self,$t);
-    } elsif (exists $t->{prechild} || exists $t->{prechildren} ||
-	     exists $t->{postchild} || exists $t->{postchildren}) {
-	warn "XML::XPathScript::Toys::call_template: cannot handle this sort of templates yet";
-	# Attempt to recover
-	$t->{pre}="";
-	$t->{post}="";
-	return 1;
-    } else {
-	$t->{pre}=$template->{pre};
-	$t->{post}=$template->{post};
-	return 1;
-    };
-}
-
-
-sub _apply_templates {
-    my @nodes = @_;
-
-    my $retval = '';
-    foreach my $node (@nodes) {
-        $retval .= translate_node($node);
-    }
-
-    return $retval;
-}
-
-sub translate_node {
-    my $node = shift;
-
-    my $translations = $XML::XPathScript::current->{t};
-
-    # Find appropriate template for node.
-    my $trans;
-    if ($node->isElementNode) {
-	my $node_name = $node->getName;
-	$trans = $translations->{$node_name} if (defined $node_name);
-	$trans ||= $translations->{'*'};
-    } elsif ($node->isTextNode) {
-	$trans = $translations->{"text()"};
-    } elsif ($node->isCommentNode) {
-	$trans = $translations->{"comment()"};
-    } elsif ($node->isPINode) {
-	unless (defined($trans=$translations->{"processing-instruction()"})) {
-	    # Default handling of Processing Instructions:
-	    # don't output top-level ones, but leave children alone
-	    # (falling thru the no-template case).
-	    eval { $node->getParentNode->getParentNode } || return '';
-	};
-    } else { # Unknown node type, leave it alone.
-        return $node->toString;
-    }
-
-    if (!$trans) { # Node has no template.
-	if ($node->isElementNode) {
-	    return start_tag($node) .
-	      _apply_templates($node->getChildNodes) .
-                end_tag($node);
-	} else {
-	    return $node->toString;
-	}
-    }
-
-
-    my $dokids = 1;
-    my $search;
-
-    if ($trans->{testcode}) {
-	$trans={%$trans}; # Performs shallow copy...
-	my $t={};         # for modifying it in place according to how $t is
-	# set by testcode sub.
-
-#         warn "Evalling testcode\n";
-        my $result = $trans->{testcode}->($node, $t);
-
-	if ($result !~ m/^-?\d+/) {
-            $dokids = 0;
-            $search = $result;
-	}
-	elsif ($result == DO_NOT_PROCESS) {
-            # don't process anything.
-            return;
-        }
-	elsif ($result == DO_SELF_ONLY) {
-            # -1 means don't do children.
-            $dokids = 0;
-        }
-        elsif ($result == DO_SELF_AND_KIDS) {
-            # do kids
-        } else {
-	    use Carp;
-	    Carp::croak "Unknown return value in testcode: $result";
-	}
-
-	if (%$t) {
-	    foreach my $key (keys %$t) {
-		$trans->{$key} = $t->{$key};
-	    }
-	}
-    }
-
-    # default: process children too.
-    my $pre = interpolate($node, $trans->{pre}) .
-            ($trans->{showtag} ? start_tag($node) : '') .
-            interpolate($node, $trans->{prechildren});
-
-    my $post = interpolate($node, $trans->{postchildren}) .
-            ($trans->{showtag} ? end_tag($node) : '') .
-            interpolate($node, $trans->{post});
-
-    if ($dokids) {
-        my $middle = '';
-        for my $kid ($node->getChildNodes()) {
-            if ($kid->isElementNode) {
-                $middle .= interpolate($node, $trans->{prechild}) .
-                        _apply_templates($kid) .
-                        interpolate($node, $trans->{postchild});
-            }
-            else {
-                $middle .= _apply_templates($kid);
-            }
-        }
-        return $pre . $middle . $post;
-    }
-    elsif ($search) {
-        my $middle = '';
-        for my $kid (findnodes($search, $node)) {
-            if ($kid->isElementNode) {
-                $middle .= interpolate($node, $trans->{prechild}) .
-                        _apply_templates($kid) .
-                        interpolate($node, $trans->{postchild});
-            }
-            else {
-                $middle .= _apply_templates($kid);
-            }
-        }
-        return $pre . $middle . $post;
-    }
-    else {
-        return $pre . $post;
-    }
-}
-
-sub start_tag {
-    my ($node) = @_;
-
-    my $name = $node->getName;
-    return '' unless $name;
-
-    my $string = "<" . $name;
-
-    foreach my $ns ($node->getNamespaceNodes) {
-        $string .= $ns->toString;
-    }
-
-    foreach my $attr ($node->getAttributeNodes) {
-        $string .= $attr->toString;
-    }
-
-    $string .= ">";
-
-    return $string;
-}
-
-sub end_tag {
-    my ($node) = @_;
-
-    if (my $name = $node->getName) {
-        return "</" . $name . ">";
-    }
-    else {
-        return '';
-    }
-}
-
-sub interpolate {
-    my ($node, $string) = @_;
-    return '' if (!defined $string);
-    return $string if (! $XML::XPathScript::current->interpolating());
-    my $new = '';
-    while ($string =~ m/\G(.*?)\{(.*?)\}/gcs) {
-        my ($pre, $path) = ($1, $2);
-        $new .= $pre;
-        $new .= $node->findvalue($path);
-    }
-    $string =~ /\G(.*)/gcs;
-    $new .= $1 if defined $1;
-    return $new;
-}
-
-1;
-
 
 =pod "
 
@@ -578,21 +171,82 @@ the same as the above in C (and hence an order of magnitude faster).
 
 =back
 
+=head2 Global variables
+
+Due to the peculiar syntax allowed in the embedded dialect for
+accessing the template hash, the stylesheet is not reentrant and
+cannot (yet) transform several documents at once. However, one should
+not rely on those variables existing forever.
+
+=over
+
+=item I<$XML::XPathScript::xp>
+
+The XML::XPath object that holds the whole document (created by
+L<XML::XPath/new>)
+
+=item I<$XML::XPathScript::trans>
+
+The template hash currently in use (known as $t in the AxKit
+documentation). Its keys are element names, and its values are
+the matching templates (as hash references).
+
+=back
+
 =cut "
 
-package XML::XPathScript;
 
-use vars qw($VERSION);
+use vars qw( $VERSION $XML_parser $DoNotInterpolate $debug_level );
 
-use XML::XPath 1.0;
-use XML::XPath::XMLParser;
-use XML::XPath::Node;
-use XML::XPath::NodeSet;
-use XML::Parser;
 use Symbol;
 use File::Basename;
+use XML::XPathScript::Toys;
 
-$VERSION = '0.11';
+$VERSION = '0.12';
+
+$XML_parser = 'XML::XPath';
+
+# By default, we interpolate
+$DoNotInterpolate = 0;
+
+# internal variable for debugging information. 
+# 0 is total silence and 10 is complete verbiage
+$debug_level = 0;
+
+sub import
+{
+	my $self = shift @_;
+
+	if( grep $_ eq 'XML::XPath', @_ )
+	{
+		$XML::XPathScript::XML_parser = 'XML::XPath';
+	}
+	elsif( grep $_ eq 'XML::LibXML', @_ )
+	{
+		$XML::XPathScript::XML_parser = 'XML::LibXML';
+	}
+}
+
+INIT
+{
+	if( $XML_parser eq 'XML::XPath' )
+	{
+		eval <<EOT;
+			use XML::XPath 1.0;
+			use XML::XPath::XMLParser;
+			use XML::XPath::Node;
+			use XML::XPath::NodeSet;
+			use XML::Parser;
+EOT
+		die $@ if $@;
+	}
+	else
+	{
+		eval 'use XML::LibXML';
+		die $@ if $@;
+	}
+
+}
 
 =pod "
 
@@ -688,26 +342,106 @@ sentence).
 sub process {
     my ($self, $printer, @extravars) = @_;
 
-    my ($xdom);
-    if (UNIVERSAL::isa($self->{xml},"XML::XPath")) {
-	$xdom=$self->{xml};
-    } elsif (UNIVERSAL::isa($self->{xml},"XML::libXML")) {
-	# Transcode to XML::XPath - OUCH ! Huge performance hit.
-	# In the future we will be able to manage both kinds of
-	# XML trees at once.
-	$xdom=XML::XPath->new(xml => $self->{xml}->toString);
-    } elsif (ref($self->{xml})) {
-	$xdom=XML::XPath->new(ioref => $self->{xml})
-    } else {
-	$xdom=XML::XPath->new(xml => $self->{xml});
-    };
-    $self->{xp}=$xdom;
-
     do { $$printer="" } if (UNIVERSAL::isa($printer, "SCALAR"));
-    $self->{printer}=$printer;
+    $self->{printer}=$printer if $printer;
 
-    $self->compile()->($self, @extravars);
 
+    my $xpath;
+
+	#warn "Entering process...";
+
+	if( $XML_parser eq 'XML::XPath' )
+	{
+		eval <<EOT;
+			use XML::XPath 1.0;
+			use XML::XPath::XMLParser;
+			use XML::XPath::Node;
+			use XML::XPath::NodeSet;
+			use XML::Parser;
+EOT
+		die $@ if $@;
+	}
+	else
+	{
+		eval 'use XML::LibXML';
+		die $@ if $@;
+	}
+
+	#warn "Eval'uation done...";
+
+
+	# a third option should be auto, for which we
+	# would use the already-defined object
+	if( $XML_parser eq 'auto' )
+	{
+		if (UNIVERSAL::isa($self->{xml},"XML::XPath")) 
+		{
+			$xpath=$self->{xml};
+			$XML_parser = 'XML::XPath';
+		}
+		elsif(UNIVERSAL::isa($self->{xml},"XML::LibXML" ))
+		{
+			$xpath=$self->{xml};
+			$XML_parser = 'XML::LibXML';
+		}
+	}
+
+    if (UNIVERSAL::isa($self->{xml},"XML::XPath")) 
+	{
+		if( $XML_parser eq 'XML::XPath' or $XML_parser eq 'auto' )
+		{
+			$xpath=$self->{xml};
+			$XML_parser = 'XML::XPath';
+		}
+		else 		# parser if XML::LibXML
+		{
+			$xpath = XML::LibXML->parse_string( $self->{xml}->toString )->documentElement;
+		}
+    } 
+	elsif (UNIVERSAL::isa($self->{xml},"XML::libXML")) 
+	{
+		if( $XML_parser eq 'XML::LibXML' or $XML_parser eq 'auto' )
+		{
+			$xpath=$self->{xml};
+			$XML_parser = 'XML::LibXML';
+		}
+		else 		# parser if xpath
+		{
+			$xpath = new XML::XPath( xml => $self->{xml}->toString );
+		}
+    } 
+	else
+	{
+		$XML_parser = 'XML::LibXML' if $XML_parser eq 'auto';
+
+		if (ref($self->{xml})) 
+		{
+			$xpath= ( $XML_parser eq 'XML::LibXML' ) ? 
+			    XML::LibXML->new->parse_fh( $self->{xml} )->documentElement :
+				XML::XPath->new(ioref => $self->{xml})
+		} 
+		else 
+		{
+			$xpath= ( $XML_parser eq 'XML::LibXML' ) ? 
+			    XML::LibXML->new->parse_string( $self->{xml} )->documentElement :
+				XML::XPath->new( xml => $self->{xml});
+		};
+	}
+
+	$self->{dom} = $xpath;
+
+	$xpath->ownerDocument->setEncoding( "UTF-8" ) 
+		if $XML_parser eq 'XML::LibXML';
+
+	{
+		local *ORIGINAL_STDOUT;
+		*ORIGINAL_STDOUT = *STDOUT;
+   		local *STDOUT = *STDOUT;
+	   	tie *STDOUT, 'XML::XPathScript::StdoutSnatcher';
+	   	my $retval = $self->compile()->( $self, @extravars );
+	   	untie *STDOUT;
+	   	return $retval;
+	}
 }
 
 =pod "
@@ -752,7 +486,10 @@ sub extract {
     my $filename=$includestack[0] || "stylesheet";
 
     my $contents;
-    if (ref($stylesheet)) {
+
+	# $stylesheet can be a filehandler
+	# or a string
+    if( ref($stylesheet) ) {
         local $/;
         $contents = <$stylesheet>;
     }
@@ -767,12 +504,12 @@ sub extract {
         my ($text, $type) = ($1, $2);
         $line += $text =~ tr/\n//; # count \n's in text
         $text =~ s/\|/\\\|/g;
-        $script .= "\$_[0]->print(q|$text|);";
+        $script .= "print(q|$text|);";
         $script .= "\n#line $line $filename\n";
         if ($type eq '<%=') {
             $contents =~ /\G(.*?)%>/gcs || die "No terminating '%>' after line $line";
             my $perl = $1;
-            $script .= "\$_[0]->print( $perl );\n";
+            $script .= "print( $perl );\n";
             $line += $perl =~ tr/\n//;
         }
         elsif ($type eq '<!--#include') {
@@ -782,9 +519,8 @@ sub extract {
                 $params{$2} = $4 if (defined $2);
             }
 
-            if (!$params{file}) {
-                die "No matching file attribute in #include at line $line";
-            }
+			die "No matching file attribute in #include at line $line"
+				unless $params{file};
 
             $script .= $self->include_file($params{file},@includestack);
         }
@@ -797,10 +533,10 @@ sub extract {
         }
     }
 
-    if ($contents =~ /\G(.*)/gcs) {
-        my ($text) = ($1);
+    if ($contents =~ /\G(.+)/gcs) {
+        my $text = $1;
         $text =~ s/\|/\\\|/g;
-        $script .= "\$_[0]->print(q|$text|);";
+        $script .= "print(q|$text|);";
     }
 
     return $script;
@@ -840,9 +576,6 @@ alternate namespaces (e.g. ``axkit://'' URIs).
 sub include_file {
     my ($self, $filename, @includestack) = @_;
 
-    # should maybe check for circular includes here...
-#warn "INCLUDE: $filename\n";
-
     if ($filename !~ m|^\.?/|) {
 	my $reldir;
 	# We guarantee that all values we insert into @includestack begin
@@ -857,6 +590,9 @@ sub include_file {
 	};
 	$filename = "$reldir/$filename";
     }
+	
+	# are we going recursive?
+	return '' if grep $_ eq $filename, @includestack;
 
     my $sym = gensym;
     open($sym, $filename) || do {
@@ -904,47 +640,57 @@ said argument.
 sub compile {
     my ($self,@extravars) = @_;
 
-    return $self->{compiledstylesheet}
-      if (defined $self->{compiledstylesheet});
+	return $self->{compiledstylesheet}
+		if defined $self->{compiledstylesheet};
 
     my $stylesheet;
 
     if (exists $self->{stylesheet}) {
-	$stylesheet=$self->{stylesheet};
-    } elsif (exists $self->{stylesheetfile}) {
-	# This hack fails if $self->{stylesheetfile} contains
-	# double quotes.  I think we can ignore this and get
-	# away.
-	$stylesheet=qq:<!--#include file="$self->{stylesheetfile}" -->:;
-    } else {
-	die "Cannot compile without a stylesheet";
+		$stylesheet=$self->{stylesheet};
+    } 
+	elsif (exists $self->{stylesheetfile}) {
+		# This hack fails if $self->{stylesheetfile} contains
+		# double quotes.  I think we can ignore this and get
+		# away.
+		$stylesheet=qq:<!--#include file="$self->{stylesheetfile}" -->:;
+    } 
+	else {
+		die "Cannot compile without a stylesheet\n";
     };
 
     my $script = $self->extract($stylesheet);
 
     my $package=gen_package_name();
-    my $eval = join("\n",
-		    "package $package;",
-		    'no strict;',   # Don't moan on sloppyly
-		    'no warnings;', # written stylesheets
-		    "use XML::XPath::Node;",
-		    'XML::XPathScript::Toys->import;',
-		    'sub {',
-		    'my (undef,'.join('',@extravars).') = @_;',
-		    'local $XML::XPathScript::current=$_[0];',
-		    'my $t = ($XML::XPathScript::current->{t} = {});',
-		    $script,
-		    "}",
-		   );
 
+	my $extravars = join ',', @extravars;
+
+	my $eval = <<EOT;
+		    package $package;
+		    no strict;   # Don't moan on sloppyly
+		    no warnings; # written stylesheets
+			
+			use $XML_parser;  
+		    XML::XPathScript::Toys->import;
+		    sub {
+		    	my (\$self, $extravars ) = \@_;
+				local \$XML::XPathScript::current=\$self;
+		    	my \$t = \$XML::XPathScript::current->{t} = {};
+				local \$XML::XPathScript::trans = \$t; # Yes,
+				# this does the sharing! Perl is a bizarre and
+				# wonderful language.
+				local \$XML::XPathScript::xp=\$self->{dom};
+				$script
+		    }
+EOT
+
+	#warn "script ready for compil: $eval";
     local $^W;
-#warn "Compiling: $eval\n";
-    my $retval=eval $eval;
-    die $@ if (!defined $retval);
-    return ($self->{compiledstylesheet}=$retval);
-}
+	$self->debug( 10, "Compiling code:\n $eval" );
+    my $retval = eval $eval;
+    die $@ unless defined $retval;
 
-=pod "
+    return $self->{compiledstylesheet} = $retval;
+}
 
 =item I<interpolating()>
 
@@ -966,8 +712,9 @@ C<testcode>:
    $t->{'link'}{testcode} = sub {
       my ($currentnode, $t) = @_;
       my $url = findvalue('@url', $currentnode);
-      $t->{pre}="<a href=\"$url\">";
+      $t->{pre}="<a href='$url'>";
       $t->{post}='</a>';
+	  return DO_SELF_AND_KIDS();
    };
 
 Interpolation is on by default. A (now undocumented) global variable
@@ -977,16 +724,21 @@ used to change the default to off, but don't do that.
 
 sub interpolating {
     my $self=shift;
-    my $retval=(exists $self->{interpolating} ?
-		$self->{interpolating} :
-		# Obsolete, for compatibility:
-		(! $XPathScript::DoNotInterpolate));
 
     $self->{interpolating}=shift if (@_);
-    return $retval;
+
+    return exists $self->{interpolating} ?
+			$self->{interpolating} :
+		    !$XML::XPathScript::DoNotInterpolate; # Obsolete, for compatibility:
 }
 
-=pod "
+
+#  $self->debug( $level, $message )
+#	Display debugging information
+
+sub debug {
+	warn $_[2] if $_[1] <= $debug_level;
+}
 
 =item I<print($text)>
 
@@ -1002,7 +754,7 @@ sub print {
     my ($self, $text)=@_;
     my $printer=$self->{printer};
     if (!defined $printer) {
-	print $text;
+	print ORIGINAL_STDOUT $text;
     } elsif (ref($printer) eq "CODE") {
 	$printer->($text);
     } elsif (UNIVERSAL::isa($printer, "SCALAR")) {
@@ -1042,34 +794,6 @@ The functions below are not methods.
 
 =over
 
-=item I<XML::XPath::Function::document>
-
-An XPath function made available to XPath expressions in the
-stylesheet, that takes one parameter (a system filename) and returns a
-nodeset consisting of the root node of a foreign document that is
-parsed from said filename. This feature can be used to process several
-input documents at once with one stylesheet, even if their respective
-DTDs are unrelated.
-
-=cut "
-
-sub XML::XPath::Function::document {
-    my $self = shift;
-    my ($node, @params) = @_;
-    die "document: Function takes 1 parameter\n" unless @params == 1;
-
-    my $parser = XML::XPath::XMLParser->new();
-
-    my $results = XML::XPath::NodeSet->new();
-    my $newdoc;
-    my $sym = gensym;
-    my $file = $params[0];
-    open($sym, $file) || die "Cannot open document() file '$file': $!";
-    $newdoc = $parser->parse( ioref => $sym );
-    $results->push($newdoc) if $newdoc;
-    return $results;
-}
-
 
 =pod "
 
@@ -1090,15 +814,23 @@ sub gen_package_name {
 
 1;
 
+# small package to catch print statements within
+# the stylesheets
+package XML::XPathScript::StdoutSnatcher;
+
+sub TIEHANDLE { my $self = ''; bless \$self, $_[0] }
+sub PRINT {
+	my $self = shift;
+	XML::XPathScript::current()->print( @_ );
+}
+
+'end of XML::XPathScript::StdoutSnatcher' ;
+
 __END__
 
 =back
 
-=head1 BUGS
 
-Due to the peculiar syntax allowed in the embedded dialect for
-accessing the template hash, this package is not reentrant and
-thus cannot transform several documents at once.
 
 =head1 AUTHORS
 
@@ -1113,10 +845,13 @@ Apache::AxKit::Language::XPathScript by Yanick Champoux
 This is free software. You may distribute it under the same terms as
 Perl itself.
 
+=head1 SEE ALSO
+
+The XPathScript Guide at http://axkit.org/wiki/view/AxKit/XPathScriptGuide
+
 =cut
 
 # Local Variables:
 # mode:cperl
 # tab-width:8
 # End:
-
